import re
from http import HTTPStatus

from .exceptions import ParserInvalidDataError
from .utils import ensure_length

_SENTINEL = object()


class Headers(dict):
    def __setitem__(self, key, value):
        key = key.lower()
        values = super().get(key)

        if values is not None:
            values.append(value)
        else:
            values = [value]

        super().__setitem__(key, values)

    def __getitem__(self, key):
        return super().__getitem__(key.lower())

    def __delitem__(self, key, value):
        return super().__delitem__(key.lower(), value)

    def get(self, key, *args, **kwargs):
        return super().get(key.lower(), *args, **kwargs)

    def pop(self, key, *args, **kwargs):
        return super().pop(key.lower(), *args, **kwargs)

    def getone(self, key, default=None):
        try:
            return self[key][0]
        except KeyError:
            return default

    def popone(self, key, default=_SENTINEL):
        try:
            return self[key].pop(0)
        except KeyError:
            if default is _SENTINEL:
                raise
            return default


def _iter_headers(headers):
    offset = 0

    while True:
        index = headers.index(b'\r\n', offset) + 2
        data = headers[offset:index]
        offset = index

        if data == b'\r\n':
            return

        yield [item.strip() for item in data.split(b':', 1)]


def _find_headers():
    data = b''

    while True:
        data += yield
        end = data.find(b'\r\n\r\n') + 4

        if end != -1:
            return _iter_headers(data[:end]), data[end:]


class HTTPResponse:
    STATUS_LINE_REGEX = re.compile(
        r'HTTP/(?P<version>\d((?=\.)(\.\d))?) (?P<status>\d+) (?P<phrase>.+)'
    )

    def __init__(self, *, version='1.1', status, phrase, headers, body):
        self.version = version
        self.status = HTTPStatus(status)
        self.phrase = phrase
        self.headers = headers
        self.body = body

    def serialize(self):
        response = [f'HTTP/{self.version} {self.status} {self.status.phrase}']

        response.extend(f'{k}: {v}' for k, v in self.headers.items())
        response.append('\r\n')

        return b'\r\n'.join(part.encode() for part in response)

    @classmethod
    def parser(cls, protocol):
        headers, body = yield from _find_headers()

        status_line, = next(headers)

        extra = {
            'protocol': protocol,
            'status_line': status_line
        }

        match = cls.STATUS_LINE_REGEX.match(status_line.decode())

        if match is None:
            raise ParserInvalidDataError('The status line is invalid', extra)

        headers_dict = Headers()

        for key, value in headers:
            headers_dict[key] = value

        content_length = headers_dict.getone(b'content-length')

        if content_length is not None:
            content_length = int(content_length.decode())
        else:
            content_length = 0

        body = yield from ensure_length(body, content_length)

        response = cls(
            version=match.group('version'), status=int(match.group('status')),
            phrase=match.group('phrase'), headers=headers_dict, body=body[:content_length]
        )

        protocol.http_response_received(response)

        return body[content_length:]


class HTTPRequest:
    METHODS = ('GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH')

    REQUEST_LINE_REXEX = re.compile(
        rf'(?P<method>{"|".join(METHODS)}) (?P<path>.+) HTTP/(?P<version>\d((?=\.)(\.\d))?)'
    )

    def __init__(self, *, version='1.1', method, path=None, headers, body):
        self.version = version
        self.method = method
        self.path = path or '/'
        self.headers = headers
        self.body = body

    def serialize(self):
        request = [f'{self.method} {self.path} HTTP/{self.version}']

        request.extend(f'{k}: {v}' for k, v in self.headers.items())
        request.append('\r\n')

        return b'\r\n'.join(part.encode() for part in request)

    @classmethod
    def parser(cls, protocol):
        headers, body = yield from _find_headers()

        request_line, = next(headers)

        extra = {
            'protocol': protocol,
            'request_line': request_line
        }

        match = cls.REQUEST_LINE_REXEX.match(request_line.decode())

        if match is None:
            raise ParserInvalidDataError('The request line is invalid', extra)

        headers_dict = Headers()

        for key, value in headers:
            headers_dict[key] = value

        content_length = headers_dict.getone(b'content-length')

        if content_length is not None:
            content_length = int(content_length.decode())
        else:
            content_length = 0

        body = yield from ensure_length(body, content_length)

        request = cls(
            version=match.group('version'), method=match.group('method'), path=match.group('path'),
            headers=headers_dict, body=body
        )

        protocol.http_request_received(request)

        return body[content_length:]



import asyncio
import base64
import os
from http import HTTPStatus
from urllib.parse import ParseResult, urlparse, urlunparse

from .exceptions import BrokenHandshakeError, WsaioError
from .http import HTTPRequest, HTTPResponse
from .protocol import WebSocketProtocol, WebSocketProtocolState
from .websocket import WebSocketCloseCode, WebSocketFrame, WebSocketOpcode


class WebSocketClient:
    def __init__(self, loop=None):
        if loop is not None:
            self.loop = loop
        else:
            self.loop = asyncio.get_event_loop()

        self.protocol = None

        self._handshake_complete = self.loop.create_future()

    async def connection_made(self, transport):
        self.protocol.set_parser(HTTPResponse.parser(self.protocol))
        self.protocol.state = WebSocketProtocolState.HANDSHAKING

        url = urlunparse(
            ParseResult('', '', self.url.path, self.url.params, self.url.query, self.url.fragment)
        )

        request = HTTPRequest(
            method='GET',
            path=url,
            headers=self.headers,
            body=b''
        )

        await self.protocol.write(request.serialize())

    def http_response_received(self, response):
        extra = {
            'response': response,
            'protocol': self
        }

        expected_status = HTTPStatus.SWITCHING_PROTOCOLS

        if response.status is not expected_status:
            self._handshake_complete.set_exception(
                BrokenHandshakeError(
                    f'Server responsed with status code {response.status} '
                    f'({response.phrase}), need status code {expected_status} '
                    f'({expected_status.phrase}) to complete handshake. '
                    f'Closing!',
                    extra
                )
            )
            return self.protocol.close()

        connection = response.headers.getone(b'connection')

        if connection is None or connection.lower() != b'upgrade':
            self._handshake_complete.set_exception(
                BrokenHandshakeError(
                    f'Server responded with "connection: {connection}", '
                    f'need "connection: upgrade" to complete handshake. '
                    f'Closing!',
                    extra
                )
            )
            return self.protocol.close()

        upgrade = response.headers.getone(b'upgrade')

        if upgrade is None or upgrade.lower() != b'websocket':
            self._handshake_complete.set_exception(
                BrokenHandshakeError(
                    f'Server responded with "upgrade: {upgrade}", '
                    f'need "upgrade: websocket" to complete handshake. '
                    f'Closing!',
                    extra
                )
            )
            return self.protocol.close()

        self.protocol.state = WebSocketProtocolState.IDLE
        self.protocol.set_parser(WebSocketFrame.parser(self.protocol))

        self._handshake_complete.set_result(None)

        self.protocol._run_callback('ws_connected')

    def data_received(self, data):
        pass

    def connection_lost(self, exc):
        pass

    def connection_closing(self, exc):
        pass

    async def parser_failed(self, exc):
        if exc is not None:
            if self.protocol.state is WebSocketProtocolState.HANDSHAKING:
                self._handshake_complete.set_exception(exc)
            else:
                close_code = WebSocketCloseCode.NORMAL_CLOSURE

                if isinstance(exc, WsaioError):
                    close_code = exc.get_extra('close_code', WebSocketCloseCode.NORMAL_CLOSURE)

                await self.send_close(close_code, str(exc).encode())

        self.protocol.close(exc)

    def ws_connected(self):
        pass

    def ws_frame_received(self, frame):
        pass

    def ws_binary_received(self, data):
        pass

    def ws_text_received(self, data):
        pass

    def ws_ping_received(self, data):
        pass

    def ws_pong_received(self, data):
        pass

    def ws_close_received(self, code, data):
        pass

    async def send_frame(self, frame, **kwargs):
        await self.protocol.write(frame.serialize(masked=True), **kwargs)

    def send_bytes(self, data, *, opcode=WebSocketOpcode.TEXT, **kwargs):
        return self.send_frame(WebSocketFrame(opcode=opcode, data=data), **kwargs)

    def send_str(self, data, *args, **kwargs):
        return self.send_bytes(data.encode(), *args, **kwargs)

    def send_ping(self, *args, **kwrags):
        return self.send_bytes(*args, **kwrags, opcode=WebSocketOpcode.PING)

    def send_pong(self, *args, **kwrags):
        return self.send_bytes(*args, **kwrags, opcode=WebSocketOpcode.PONG)

    async def send_close(self, code, data, *, drain=True):
        data = code.to_bytes(2, 'big', signed=False) + data
        await self.send_frame(
            WebSocketFrame(opcode=WebSocketOpcode.CLOSE, data=data, drain=drain)
        )

    async def connect(self, url, *args, **kwargs):
        self.sec_ws_key = base64.b64encode(os.urandom(16))

        self.headers = kwargs.pop('headers', {})

        self.url = urlparse(url)
        self.ssl = kwargs.pop('ssl', self.url.scheme == 'wss')
        self.port = kwargs.pop('port', 443 if self.ssl else 80)

        self.headers.update({
            'Host': f'{self.url.hostname}:{self.port}',
            'Connection': 'Upgrade',
            'Upgrade': 'websocket',
            'Sec-WebSocket-Key': self.sec_ws_key.decode(),
            'Sec-WebSocket-Version': 13
        })

        kwargs.setdefault('ssl', self.ssl)

        await self.loop.create_connection(
            lambda: WebSocketProtocol(self), self.url.hostname, self.port, *args, **kwargs
        )

        await self._handshake_complete



import asyncio
import enum

from .exceptions import ConnectionClosedError


class WebSocketProtocolState(enum.IntEnum):
    INIT = 0
    IDLE = 1
    PARSING = 2
    CLOSED = 3
    HANDSHAKING = 4


class WebSocketProtocol(asyncio.Protocol):
    def __init__(self, client):
        self.client = client
        self.client.protocol = self
        self.loop = self.client.loop

        self._paused = False
        self._drain_waiter = None

        self._parser = None

        self.transport = None
        self.state = WebSocketProtocolState.INIT

        self.extensions = []

    def _run_callback(self, name, *args, **kwargs):
        func = getattr(self.client, name)
        coro = func(*args, **kwargs)

        if asyncio.iscoroutinefunction(func):
            self.loop.create_task(coro)

    def _set_paused(self, paused):
        assert self._paused is (not paused)
        self._paused = paused

    def connection_made(self, transport):
        self.transport = transport
        self._run_callback('connection_made', transport)

    def set_parser(self, parser):
        parser.send(None)
        self._parser = parser

    def data_received(self, data):
        state = self.state

        self.state = WebSocketProtocolState.PARSING

        while data:
            try:
                self._parser.send(data)
                break
            except StopIteration as err:
                data = err.value
            except Exception as err:
                self._run_callback('parser_failed', err)
                break

        self.state = state

        self._run_callback('data_received', data)

    def pause_writing(self):
        self._set_paused(True)

    def resume_writing(self):
        self._set_paused(False)

        waiter = self._drain_waiter

        if waiter is not None:
            self._drain_waiter = None

            if not waiter.done():
                waiter.set_result(None)

    def connection_lost(self, exc):
        self.state = WebSocketProtocolState.CLOSED

        if not self._paused:
            return

        waiter = self._drain_waiter

        if waiter is None:
            return

        self._drain_waiter = None

        if waiter.done():
            return

        if exc is not None:
            waiter.set_result(exc)
        else:
            waiter.set_result(None)

        self._run_callback('connection_lost', exc)

    def connection_closing(self, exc):
        self._run_callback('connection_closing', exc)

    def http_response_received(self, response):
        self._run_callback('http_response_received', response)

    def ws_connected(self):
        self._run_callback('ws_connected')

    def ws_frame_received(self, frame):
        self._run_callback('ws_frame_received', frame)

    def ws_binary_received(self, data):
        self._run_callback('ws_binary_received', data)

    def ws_text_received(self, data):
        self._run_callback('ws_text_received', data)

    def ws_ping_received(self, data):
        self._run_callback('ws_ping_received', data)

    def ws_pong_received(self, data):
        self._run_callback('ws_pong_received', data)

    def ws_close_received(self, code, data):
        self._run_callback('ws_close_received', code, data)

    async def drain(self):
        if self.state is WebSocketProtocolState.CLOSED:
            raise ConnectionClosedError(
                'Attempt to drain a closed transport', {'protocol': self}
            )

        if not self._paused:
            return

        assert self._drain_waiter is None or self._drain_waiter.cancelled()

        waiter = self._drain_waiter = self.loop.create_future()

        return waiter

    async def write(self, data, *, wait=False):
        if self.state is WebSocketProtocolState.CLOSED:
            raise ConnectionClosedError(
                'Attempt to write to a closed transport', {'procotol': self}
            )

        self.transport.write(data)

        if wait:
            await self.drain()

    def close(self, exc=None):
        self.state = WebSocketProtocolState.CLOSED
        self._run_callback('connection_closing', exc)

        if self.transport is not None:
            self.transport.close()
